"""Contains the main function."""

import atexit
import logging
import time
import threading
import signal
import sys
import notify2
import requests
from twitchgamenotify.cache import (
    is_cache_locked,
    load_cache,
    lock_cache,
    unlock_cache,
    save_cache,
)
from twitchgamenotify.constants import STARTING_CONNECTION_FAILURE_RETRY_TIME
from twitchgamenotify.configuration import (
    ConfigFileNotFound,
    parse_config_file,
    parse_runtime_args,
)
from twitchgamenotify.notifications import (
    process_notifications_wrapper,
    send_connection_error_notification,
)
from twitchgamenotify.twitch_api import TwitchApi
from twitchgamenotify.version import NAME


def graceful_exit(*_, **__):
    """Exit such that atexit gets triggered."""
    logging.info("Exitting %s", NAME)
    sys.exit(0)


def main():
    """The main function."""
    # Get runtime arguments
    cli_args = parse_runtime_args()

    # Set up logger
    logging.basicConfig(
        format="%(levelname)s: %(message)s", level=cli_args.loglevel
    )

    # Make sure process kills and poweroffs cause atexit registers to
    # trigger
    signal.signal(signal.SIGTERM, graceful_exit)
    signal.signal(signal.SIGINT, graceful_exit)

    # Load cached static API data
    if not cli_args.no_caching:
        # Load the cache
        cache_dict = load_cache()

        if is_cache_locked():
            # Another process has claimed the cache
            logging.warning(
                "Cache file is locked. All static API data "
                "collected during this run will be discarded."
            )
        else:
            # Claim ownership of the cache
            lock_cache()

            # Create a hook to save and unlock the cache when exiting
            # the program
            atexit.register(unlock_cache)
            atexit.register(save_cache, cache_dict)

    # Read config file
    try:
        config_dict = parse_config_file()
    except ConfigFileNotFound:
        logging.error("Config file not found. Aborting.")
        sys.exit(1)

    # Set up the notifier
    if not cli_args.print_to_terminal:
        notify2.init(NAME)

    # Set up app indicator and run it in a separate thread
    if (
        not cli_args.one_shot
        or not cli_args.print_to_terminal
        or cli_args.no_app_indicator
    ):
        # Import this here so GTK-incompatible machines are still
        # supported
        from twitchgamenotify.app_indicator import AppIndicator

        indicator = AppIndicator()

        # Kill the indicator when we quit
        atexit.register(indicator.stop)

        # Start the indicator in its own thread
        threading.Thread(target=indicator.start, daemon=True).start()

    # Connect to the API - keep retrying and be loud if there's a
    # connection error
    while True:
        try:
            twitch_api = TwitchApi(
                client_id=config_dict["twitch-api-client-id"],
                client_secret=config_dict["twitch-api-client-secret"],
            )

            # Successful connection
            break
        except requests.exceptions.ConnectionError:
            # Internet is probably down. Log an error and notify if we're
            # notifying
            send_connection_error_notification(
                send_dbus_notification=not cli_args.print_to_terminal
            )

            logging.info(
                "Retrying in %s seconds",
                STARTING_CONNECTION_FAILURE_RETRY_TIME,
            )

            # Wait a bit before retrying
            time.sleep(STARTING_CONNECTION_FAILURE_RETRY_TIME)

    # Set up arguments to give process_notifcations
    kwargs = dict(
        print_to_terminal=cli_args.print_to_terminal,
        streamers=config_dict["streamers"],
        twitch_api=twitch_api,
    )

    if not cli_args.no_caching:
        kwargs["names_cache"] = cache_dict

    if not cli_args.one_shot:
        # Remember what game a streamer was playing last so we don't
        # re-notify
        streamers_last_seen_playing_dict = {
            streamer: "" for streamer in config_dict["streamers"].keys()
        }

        kwargs["streamers_previous_game"] = streamers_last_seen_playing_dict

    # Query (and possibly notify) only once or periodically
    if cli_args.one_shot:
        process_notifications_wrapper(**kwargs)
    else:
        # Loop until we get interrupted
        while True:
            # Process any notifications
            threading.Thread(
                target=process_notifications_wrapper,
                kwargs=kwargs,
                daemon=True,
            ).start()

            # Wait before querying again
            time.sleep(config_dict["query-period"])



"""Contains an application indicator.

Thanks to Timur Rubeko with his wonderful guide here:
https://candidtim.github.io/appindicator/2014/09/13/ubuntu-appindicator-step-by-step.html.
"""

import _thread  # pylint: disable=wrong-import-order
import logging
import gi
from twitchgamenotify.constants import APP_INDICATOR_SVG_PATH
from twitchgamenotify.version import NAME

# Make sure our GTK stuff is good before loading the app indicator
try:
    gi.require_version("Gtk", "3.0")
    gi.require_version("AppIndicator3", "0.1")
except ValueError as e:
    logging.error(e)
    _thread.interrupt_main()

from gi.repository import (
    Gtk,
    AppIndicator3,
)  # pylint: disable=wrong-import-order,wrong-import-position,ungrouped-imports


class AppIndicator:
    """Shows a simple app indicator for the program."""

    def __init__(self):
        """Setup the app indicator."""
        # Setup
        self.indicator = AppIndicator3.Indicator.new(
            NAME,
            APP_INDICATOR_SVG_PATH,
            AppIndicator3.IndicatorCategory.SYSTEM_SERVICES,
        )
        self.indicator.set_status(AppIndicator3.IndicatorStatus.ACTIVE)

        # Build the menu
        menu = Gtk.Menu()
        item_quit = Gtk.MenuItem("Quit")

        # Wrap the kill_main function
        quit_wrapper = lambda source: _thread.interrupt_main()

        item_quit.connect("activate", quit_wrapper)
        menu.append(item_quit)
        menu.show_all()

        self.indicator.set_menu(menu)

    @staticmethod
    def start():
        """Starts the Gtk main loop (and hence the indicator)."""
        Gtk.main()

    @staticmethod
    def stop():
        """Stops the Gtk main loop (and hence the indicator)."""
        Gtk.main_quit()



"""Functions for caching static API data.

For all of these functions, the cache is located at
$XDG_CONFIG_HOME/twitch-game-notify/cache.json, where XDG_CONFIG_HOME
defaults to $HOME/.config.
"""

import json
import os
from twitchgamenotify.constants import (
    CACHE_FILE_LOCK_NAME,
    CACHE_FILE_NAME,
    PROJECT_CONFIG_HOME,
)


def is_cache_locked():
    """Determines whether a cache is locked.

    Returns:
        A boolean signalling whether the cache is locked.
    """
    # Build the path to the cache lock
    cache_lock_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_LOCK_NAME)

    return os.path.exists(cache_lock_path)


def lock_cache():
    """Creates a cache lock file."""
    # Build the path to the cache lock
    cache_lock_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_LOCK_NAME)

    # Create an empty lock file
    open(cache_lock_path, "x")


def unlock_cache(catch_failure=True):
    """Removes a cache lock file.

    Arg:
        catch_failure: An optional boolean signalling whether to catch
            a FileNotFoundError exception occuring when the cache lock
            to remove doesn't exist.
    """
    # Build the path to the cache lock
    cache_lock_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_LOCK_NAME)

    # Remove the lock
    if catch_failure:
        try:
            os.remove(cache_lock_path)
        except FileNotFoundError:
            pass
    else:
        os.remove(cache_lock_path)


def load_cache():
    """Loads a cache, or creates one if it doesn't exit.

    Returns:
        A dictionary containing static API data. For example:

        {"games": {"460630": "Tom Clancy's Rainbow Six: Siege",
                   "493057": "PLAYERUNKNOWN'S BATTLEGROUNDS"},
         "streamers": {"macie_jay": "Macie_Jay",
                       "shroud": "shroud"}}
    """
    # Build the path to the cache JSON file
    cache_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_NAME)

    # If a cache doesn't exist or if it's an empty file (maybe it got
    # corrupted at some point?), return an empty dictionary as the new
    # cache
    if not os.path.exists(cache_path) or not os.stat(cache_path).st_size:
        return {"games": {}, "streamers": {}}

    # Load the cache file
    with open(cache_path, "r") as cache_file:
        return json.load(cache_file)


def save_cache(cache_dictionary):
    """Saves a cache as JSON.

    Arg:
        cache_dictionary: A dictionary containing cache information to
            be serialized to JSON.
    """
    # Get the path for the cache
    cache_path = os.path.join(PROJECT_CONFIG_HOME, CACHE_FILE_NAME)

    # Make sure the directory paths to the cache file exist
    os.makedirs(PROJECT_CONFIG_HOME, exist_ok=True)

    with open(cache_path, "w") as cache_file:
        cache_file.write(
            json.dumps(cache_dictionary, indent=2, sort_keys=True)
        )



"""Stores constants for all modules."""

import logging
import os


# Base of the repository/project
PROJECT_BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Path to the app indicator icon
APP_INDICATOR_SVG_PATH = os.path.join(
    PROJECT_BASE_DIR, "static/", "twitch.svg"
)


# Base of XDG config files
try:
    PROJECT_CONFIG_HOME = os.path.join(
        os.environ["XDG_CONFIG_HOME"], "twitch-game-notify"
    )
except KeyError:
    PROJECT_CONFIG_HOME = os.path.join(
        os.environ["HOME"], ".config/", "twitch-game-notify"
    )


# Config file names
CONFIG_FILE_NAME = "config.yaml"
CACHE_FILE_NAME = "cache.json"
CACHE_FILE_LOCK_NAME = CACHE_FILE_NAME + ".lock"


# URL for example config file on Github
EXAMPLE_CONFIG_FILE_URL = "https://raw.githubusercontent.com/mwiens91/twitch-game-notify/master/config.yaml.example"


# Loglevel CLI options
CRITICAL = "critical"
ERROR = "error"
WARNING = "warning"
INFO = "info"
DEBUG = "debug"

LOGLEVEL_CHOICES = [CRITICAL, ERROR, WARNING, INFO, DEBUG]
LOGLEVEL_DICT = {
    CRITICAL: logging.CRITICAL,
    ERROR: logging.ERROR,
    WARNING: logging.WARNING,
    INFO: logging.INFO,
    DEBUG: logging.DEBUG,
}


# Twitch API URLs
TWITCH_BASE_API_URL = "https://api.twitch.tv/helix"
TWITCH_STREAM_API_URL = TWITCH_BASE_API_URL + "/streams"
TWITCH_USER_API_URL = TWITCH_BASE_API_URL + "/users"
TWITCH_GAME_API_URL = TWITCH_BASE_API_URL + "/games"
TWITCH_TOKEN_API_URL = "https://id.twitch.tv/oauth2/token"


# HTTP status codes
HTTP_200_OK = 200
HTTP_401_UNAUTHORIZED = 401


# How many seconds to wait when initially no connection
STARTING_CONNECTION_FAILURE_RETRY_TIME = 10



"""Functions for processing and displaying notifications."""

import datetime
import logging
import time
import notify2
import requests
from twitchgamenotify.twitch_api import FailedHttpRequest
from twitchgamenotify.version import NAME


# ANSI escape sequence for bold text
ANSI_BOLD = "\033[1m"
ANSI_END = "\033[0m"


def print_notification_to_terminal(streamer_name, stream_title, game_name):
    """Print a game notification to the terminal.

    Args:
        streamer_name: A string containing the name of the streamer.
        stream_title: A string containing the title of the stream.
        game_name: A string containing the name of the game.
    """
    print(ANSI_BOLD + streamer_name + ANSI_END, end="")
    print(" @ " + datetime.datetime.now().isoformat())
    print("Title: %s" % stream_title)
    print("Playing: %s" % game_name)


def send_notification_to_dbus(streamer_name, stream_title, game_name):
    """Send a game notification to D-Bus.

    Args:
        streamer_name: A string containing the name of the streamer.
        stream_title: A string containing the title of the stream.
        game_name: A string containing the name of the game.
    """
    notify2.Notification(
        streamer_name + " @ " + time.strftime("%H:%M"),
        "Title: %s\nPlaying: %s" % (stream_title, game_name),
    ).show()


def send_error_notification(error_message, send_dbus_notification):
    """Logs and notifies an error message.

    Args:
        error_message: A string containing the error message
        send_dbus_notification: A boolean specifying whether to send a
            notification to D-Bus.
    """
    # Log the error
    logging.error(error_message)

    # Send a notification about the error, if instructed to
    if send_dbus_notification:
        notify2.Notification(
            NAME + " @ " + time.strftime("%H:%M"), error_message
        ).show()


def process_notifications(
    streamers,
    twitch_api,
    names_cache=None,
    streamers_previous_game=None,
    print_to_terminal=False,
):
    """Query the Twitch API for all streamers and display notications.

    Args:
        names_cache: An optional dictionary containing cached names for
            games and streamers. For example:

            {"games": {"460630": "Tom Clancy's Rainbow Six: Siege"},
             "streamers": {"macie_jay": "Macie_Jay"}}

            Defaults to None.
        print_to_terminal: An optional boolean signalling whether to
            print to the terminal instead of passing a message to D-Bus.
            Defaults to False.
        streamers: A dictionary of streamers where the keys are strings
            containing the streamer's login name and the values are
            dictionaries containing the user's settings for the
            streamer. For example:

            {'macie_jay': {'include': '460630'},
             'moonmoon_ow': {'include': '*', 'exclude': '33214'}}

        streamers_previous_game: An optional dictionary containing
            information about what game a streamer was last seen
            playing.  The keys are strings containing the streamers
            login name, and the keys are strings containing the game ID
            of what they were last seen playing (or an empty string if
            the streamer hasn't yet been seen live). This defaults to
            None, which is useful if this function is only being called
            once.
        twitch_api: An authenticated TwitchApi object to interact with
            Twitch's API.
    """
    # Look up info about each streamer's stream
    for streamer_login_name, games in streamers.items():
        try:
            # Get info about stream
            info = twitch_api.get_online_stream_info(streamer_login_name)
        except FailedHttpRequest as e:
            # Bad HTTP request! Log the error and move onto the next
            # streamer!
            send_error_notification(e.message, not print_to_terminal)

            continue

        # If the streamer isn't live, record that they aren't playing
        # anything and move onto the next streamer
        if not info["live"]:
            # Mark them as last seen playing nothing
            if (
                streamers_previous_game
                and streamers_previous_game[streamer_login_name]
            ):
                streamers_previous_game[streamer_login_name] = ""

            continue

        # Check if this is a game to notify about
        game_id = info["game_id"]

        # If the streamer was last seen playing this game, move on. If
        # they are playing something new, record it.
        if streamers_previous_game:
            if streamers_previous_game[streamer_login_name] == game_id:
                # The streamer is playing the same game as before
                continue

            # Streamer is playing something new. Update the previously
            # seen game.
            streamers_previous_game[streamer_login_name] = game_id

        # Check the include (and possibly exclude) list
        if "*" in games["include"]:
            # All games are included. Check if we need to exclude any
            # games.
            if "exclude" in games and game_id in games["exclude"]:
                continue

        elif game_id not in games["include"]:
            # Game not in the include list
            continue

        # Gather info about the stream. Lookup info in the cache first
        # if it's available.
        if names_cache is not None:
            # Try getting the display name from the cache
            try:
                streamer_display_name = names_cache["streamers"][
                    streamer_login_name
                ]
            except KeyError:
                # Fetch it
                try:
                    streamer_display_name = twitch_api.get_streamer_display_name(
                        streamer_login_name
                    )
                except FailedHttpRequest as e:
                    # Bad HTTP request! Log the error and move onto the next
                    # streamer!
                    send_error_notification(e.message, not print_to_terminal)

                    continue

                # Store it
                names_cache["streamers"][
                    streamer_login_name
                ] = streamer_display_name

            # Try getting the game title from the cache
            try:
                game_title = names_cache["games"][game_id]
            except KeyError:
                # Fetch it
                try:
                    game_title = twitch_api.get_game_title(game_id)
                except FailedHttpRequest as e:
                    # Bad HTTP request! Log the error and move onto the next
                    # streamer!
                    send_error_notification(e.message, not print_to_terminal)

                    continue

                # Store it
                names_cache["games"][game_id] = game_title
        else:
            # Not using cache. Fetch everything.
            try:
                streamer_display_name = twitch_api.get_streamer_display_name(
                    streamer_login_name
                )
                game_title = twitch_api.get_game_title(game_id)
            except FailedHttpRequest as e:
                # Bad HTTP request! Log the error and move onto the next
                # streamer!
                send_error_notification(e.message, not print_to_terminal)

                continue

        # Title is never cached
        stream_title = info["title"]

        # Send a notification
        if print_to_terminal:
            print_notification_to_terminal(
                streamer_display_name, stream_title, game_title
            )
        else:
            send_notification_to_dbus(
                streamer_display_name, stream_title, game_title
            )


def send_connection_error_notification(send_dbus_notification):
    """Logs and notifies about a connection failure.

    Arg:
        send_dbus_notification: A boolean specifying whether to send a
            notification to D-Bus.
    """
    # The message to show
    error_message = "Unable to connect to Twitch. Is your internet okay?"

    # Show the message
    send_error_notification(error_message, send_dbus_notification)


def process_notifications_wrapper(*args, **kwargs):
    """A wrapper for process_notifications to catch connection errors.

    This makes the code a bit cleaner than inserting lots of try/except
    blocks into the process_notifications function.
    """
    # Determine whether to print to the terminal
    try:
        display_dbus_notification = not kwargs["print_to_terminal"]
    except KeyError:
        # In accordance with print_to_terminal defaulting to False
        display_dbus_notification = True

    # Try calling process_notifications
    try:
        process_notifications(*args, **kwargs)
    except requests.exceptions.ConnectionError:
        # Bad connection - stop this iteration
        send_connection_error_notification(
            send_dbus_notification=display_dbus_notification
        )



#!/usr/bin/env python3

from setuptools import setup
from twitchgamenotify.version import NAME, DESCRIPTION, VERSION


# Parse readme to include in PyPI page
with open("README.md") as f:
    long_description = f.read()


def capitalize(s):
    """Capitalize the first letter of a string.

    Unlike the capitalize string method, this leaves the other
    characters untouched.
    """
    return s[:1].upper() + s[1:]


setup(
    name=NAME,
    version=VERSION,
    description=capitalize(DESCRIPTION),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mwiens91/twitch-game-notify",
    author="Matt Wiens",
    author_email="mwiens91@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
    packages=["twitchgamenotify", "static"],
    package_data={"static": ["twitch.svg"]},
    entry_points={
        "console_scripts": ["twitch-game-notify = twitchgamenotify.main:main"]
    },
    python_requires=">=3.5",
    install_requires=[
        "dbus-python",
        "notify2",
        "PyGObject",
        "PyYAML",
        "requests",
    ],
)



"""Contains configuration related functions."""

import argparse
import os.path
import sys
import requests
import yaml
from twitchgamenotify.cache import unlock_cache
from twitchgamenotify.constants import (
    CONFIG_FILE_NAME,
    EXAMPLE_CONFIG_FILE_URL,
    LOGLEVEL_CHOICES,
    LOGLEVEL_DICT,
    PROJECT_BASE_DIR,
    PROJECT_CONFIG_HOME,
    WARNING,
)
from twitchgamenotify.version import NAME, VERSION, DESCRIPTION


class ConfigFileNotFound(Exception):
    """Raised when a config file can't be found."""

    pass


class PrintExampleConfigAction(argparse.Action):
    """argparse action to print example config file."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Download and print example config file."""
        r = requests.get(EXAMPLE_CONFIG_FILE_URL)

        if r.status_code != 200:
            print(
                "ERROR: Failed to download example config file!",
                file=sys.stderr,
            )
            sys.exit(1)
        else:
            print(r.content.decode())
            sys.exit(0)


class RemoveCacheLockAction(argparse.Action):
    """argparse action to remove cache lock (should one exist)."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Remove the cache lock."""
        try:
            unlock_cache(catch_failure=False)
        except FileNotFoundError:
            print(
                "WARNING: No existing cache lock to remove!", file=sys.stderr
            )

        # Get out
        sys.exit(0)


class TranslateLogLevelAction(argparse.Action):
    """argparse action to translate loglevel to a number."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Set a logging loglevel number."""
        setattr(namespace, self.dest, LOGLEVEL_DICT[values])


def find_config_file():
    """Find and return the path of a config file.

    The config file looked for is "config.yaml" and it is looked for at
    the base of the respository first (if you're running from source),
    and then in $XDG_CONFIG_HOME/twitch-game-notify/ (XDG_CONFIG_HOME
    defaults to $HOME/.config).

    Returns:
        A string containing the absolute path to the config file.

    Raises:
        ConfigFileNotFound: A config file couldn't be found.
    """
    # Check the base of the project
    config_path = os.path.join(PROJECT_BASE_DIR, CONFIG_FILE_NAME)

    if os.path.exists(config_path):
        return config_path

    # Check XDG_CONFIG_HOME
    config_path = os.path.join(PROJECT_CONFIG_HOME, CONFIG_FILE_NAME)

    if os.path.exists(config_path):
        return config_path

    # Couldn't find anything :thinking:
    raise ConfigFileNotFound


def parse_config_file():
    """Find and parse a config file."""
    # Find the config file first
    config_path = find_config_file()

    # Now parse and return it - note that PyYAML doesn't come with any
    # schema validation, which might be desirable at some point
    with open(config_path, "r") as config_file:
        return yaml.safe_load(config_file)


def parse_runtime_args():
    """Parse runtime args using argparse.

    Returns:
        An object of type 'argparse.Namespace' containing the runtime
        arguments as attributes. See argparse documentation for more
        details.
    """
    parser = argparse.ArgumentParser(
        prog=NAME, description="%(prog)s - " + DESCRIPTION
    )
    parser.add_argument(
        "-l",
        "--loglevel",
        default=LOGLEVEL_DICT[WARNING],
        choices=LOGLEVEL_CHOICES,
        action=TranslateLogLevelAction,
        help="how much to log",
    )
    parser.add_argument(
        "--one-shot", action="store_true", help="query once then exit"
    )
    parser.add_argument(
        "--no-app-indicator",
        action="store_true",
        help="don't launch an app indicator",
    )
    parser.add_argument(
        "--no-caching", action="store_true", help="don't cache static API data"
    )
    parser.add_argument(
        "--print-config",
        nargs=0,
        action=PrintExampleConfigAction,
        help="downloads and prints example config file",
    )
    parser.add_argument(
        "--print-to-terminal",
        action="store_true",
        help="print to terminal (doesn't connect to D-Bus)",
    )
    parser.add_argument(
        "--remove-cache-lock",
        nargs=0,
        action=RemoveCacheLockAction,
        help="removes a cache lock if one exists",
    )
    parser.add_argument(
        "--version", action="version", version="%(prog)s " + VERSION
    )

    return parser.parse_args()
